from typing import Union, Iterable, Optional, NamedTuple
import numpy as np
from numpy.lib import recfunctions as rfn

from abito.lib.stats import weighted, plain
from abito.lib.significance import *
from abito.lib.significance import _t_test_from_stats
from abito.lib.bootstrap import generate_bootstrap_estimates
from abito.lib.prepare.trim import argtrim
from abito.lib.prepare.compress import reweigh, compress, _validate_compress_stat
from abito.lib.utils import _return_or_inplace


__all__ = ['sample', 'Sample', 'LinearSample', 'RatioSample']


class SampleBase:
    """
    Base class with shared methods
    """
    def __init__(self, obs: np.ndarray, weights: np.ndarray = np.empty(0, dtype='int')):
        """
        Parameters
        ----------
        obs : ndarray
            If RatioSample is being initialized, `obs` should be structured array with field names ('num', 'den')
        weights : ndarray, optional
            Weights. If not specified `weights` assigned to empty ndarray
        """
        self.obs = obs
        self.weights = weights

    def copy(self):
        """
        Return copy of the object.

        Returns
        -------
        LinearSample or RatioSample
        """
        return self.__class__(self.obs.copy(), self.weights.copy())

    @property
    def is_weighted(self) -> bool:
        """
        Is Sample has weights.
        Returns
        -------
        bool
        """
        return self.weights.shape[0] > 0

    @property
    def fullobs(self) -> np.ndarray:
        """
        Full observations vector after weights applied

        Returns
        -------
        out : ndarray

        """
        if self.is_weighted:
            return np.repeat(self.obs, self.weights)
        else:
            return self.obs

    @property
    def _stat_funcs_module(self):
        if self.is_weighted:
            return weighted
        else:
            return plain

    @property
    def nobs(self) -> np.int:
        """
        Number of observations in sample.

        Returns
        -------
        out : int

        """
        if self.is_weighted:
            return self.weights.sum()
        else:
            return self.obs.shape[0]

    def reweigh(self, inplace=False):
        """
        Make sample weighted. Initialize weights by grouping unique values of observations.
        Reinitialise if sample already has weights (may be used in case of duplicates).

        Parameters
        ----------
        inplace : bool, optional
            If ``True``, reinitialize obs / weights of the instance, default is ``False``.

        Returns
        -------
        out : LinearSample or RatioSample
            If `inplace` is ``True``, returns None.
        """
        obs, weights = reweigh(self.obs, self.weights)
        return _return_or_inplace(self, obs, weights, inplace)

    @property
    def cache(self):
        """
        Dict property that stores results of bootstrap_estimates.

        Returns
        -------
        out : dict of ndarrays
        """
        if not hasattr(self, '_cache'):
            self.cache = {}
        return self._cache

    @cache.setter
    def cache(self, cache):
        self._cache = cache

    @cache.deleter
    def cache(self):
        self._cache = {}

    @staticmethod
    def _build_cache_key(**key_args):
        key_args = {k: tuple(v) if isinstance(v, Iterable) else v for k, v in key_args.items()}
        return tuple(sorted(key_args.items()))

    def _store_in_cache(self, value, **key_args):
        key = self._build_cache_key(**key_args)
        self.cache[key] = value

    def _get_from_cache(self, **key_args):
        key = self._build_cache_key(**key_args)
        return self.cache.get(key, None)

    def _del_from_cache(self, **key_args):
        key = self._build_cache_key(**key_args)
        if key in self.cache:
            del self.cache[key]

    def stat(self, stat: str, **stat_args) -> Union[np.float, np.int, np.ndarray]:
        """
        Calculate statistic using name and parameters.

        Parameters
        ----------
        stat : str
            {'sum', 'mean', 'var', 'std', 'mean_std', 'quantile', 'median'} for LinearSample, 'ratio' for RatioSample
        stat_args : dict
            Additional argumants that needed by corresponding method. At this time only for quantile.

        Returns
        -------
        out : number or array_like
            array_like is returned only in case of quantile and iterable `q`

        """
        stat_func = getattr(self._stat_funcs_module, stat)
        if self.is_weighted:
            return stat_func(self.obs, self.weights, **stat_args)
        else:
            return stat_func(self.obs, **stat_args)

    def bootstrap_estimates(self, stat, n_iters, n_threads=1, cache_result=False, **stat_args) -> np.ndarray:
        """

        Parameters
        ----------
        stat : str
            Name of statistics.
        n_iters : int
            Number of bootstrap iterations, size of output result.
        n_threads : int
            Number of processes to use in parallel calculation.
        cache_result : bool
            Cache results to save time in future analysis.
        stat_args : dict
            Additional arguments to pass in stat func.

        Returns
        -------
        out : array
            Array of bootstrap estimates
        """
        from_cache = self._get_from_cache(stat=stat, n_iters=n_iters, **stat_args)
        if from_cache is not None:
            return from_cache
        stat_func = getattr(self._stat_funcs_module, stat)
        res = generate_bootstrap_estimates(self.obs, stat_func, n_iters, self.weights, n_threads, **stat_args)
        if cache_result:
            self._store_in_cache(res, stat=stat, n_iters=n_iters, **stat_args)
        return res

    def bootstrap_test(
            self,
            control: 'SampleBase',
            stat: str,
            n_iters: int,
            n_threads: int = 1,
            cache_result: bool = False,
            **stat_args
    ) -> NamedTuple:
        """
        Significance test based on bootstrap estimates.

        Parameters
        ----------
        control : LinearSample or RatioSample
            Control sample.
        stat : str
            Name of statistics.
        n_iters : int
            Number of bootstrap iterations, size of output result.
        n_threads : int
            Number of processes to use in parallel calculation.
        cache_result : bool
            Cache results to save time in future analysis.
        stat_args : dict
            Additional arguments to pass in stat func.

        Returns
        -------
            estimates_diff_std : float
                Standard deviation of bootstrap estimates.
            est_p_value : float
                P-value calculated as share of estimates above / below zero.
            est_t_statistic : float
                t-statistic derived from est_p_value using inverted survival function of t-distribution.
            median_est_diff : float
                Difference between estimates' median.
            median_est_t_statistic : float
                Equal to median_est_diff / estimates_diff_std
            median_est_t_p_value : float
                P-value derived from median_est_t_statistic using survival function of t-distribution.
            stat_diff : float
                Difference between statistics if treatment and control.
            t_statistic : float
                Equal to stat_diff / estimates_diff_std.
            t_p_value : float
                P-value derived from t_statistic using survival function of t-distribution

        """
        t_stat = self.stat(stat, **stat_args)
        c_stat = control.stat(stat, **stat_args)
        t_bootstrap_estimates = self.bootstrap_estimates(stat, n_iters, n_threads, cache_result, **stat_args)
        c_bootstrap_estimates = control.bootstrap_estimates(stat, n_iters, n_threads, cache_result, **stat_args)
        return bootstrap_test(t_stat, t_bootstrap_estimates, self.nobs, c_stat, c_bootstrap_estimates, control.nobs)


class LinearSample(SampleBase):

    def compress(
            self,
            n_buckets: int,
            stat: str = 'mean',
            weights_dist: str = 'uniform',
            reweigh: bool = False,
            inplace=False,
    ):
        """

        Parameters
        ----------
        n_buckets
        stat
        weights_dist
        reweigh
        inplace

        Returns
        -------

        """
        _validate_compress_stat(stat)
        obs = self.fullobs
        obs.sort()
        compressed, weights = compress(obs, n_buckets, stat, weights_dist, reweigh)
        return _return_or_inplace(self, compressed, weights, inplace)

    def trim(self, ltrim=0, rtrim=0, inplace=False) -> Optional['LinearSample']:
        """

        Parameters
        ----------
        ltrim
        rtrim
        inplace

        Returns
        -------

        """

        trim_ind, weights = argtrim(self.obs, self.weights, ltrim=ltrim, rtrim=rtrim)

        trimmed = self.obs[trim_ind]
        samp = LinearSample(trimmed, weights)

        return _return_or_inplace(self, samp.obs, samp.weights, inplace)

    def sum(self) -> np.float:
        """

        Returns
        -------

        """
        return self.stat('sum')

    def mean(self) -> np.float:
        """

        Returns
        -------

        """
        return self.stat('mean')

    def var(self) -> np.float:
        """

        Returns
        -------

        """
        return self.stat('var')

    def std(self) -> np.float:
        """

        Returns
        -------

        """
        return self.stat('std')

    def mean_std(self) -> np.float:
        """

        Returns
        -------

        """
        return self.stat('mean_std')

    def quantile(self, q) -> np.float:
        """

        Parameters
        ----------
        q

        Returns
        -------

        """
        return self.stat('quantile', q=q)

    def median(self) -> np.float:
        """

        Returns
        -------

        """
        return self.stat('median')

    def t_test_1samp(self, popmean: float) -> NamedTuple:
        """

        Parameters
        ----------
        popmean

        Returns
        -------

        """
        return _t_test_from_stats(self.mean(), popmean, self.mean_std(), self.nobs - 1)

    def t_test(self, control: 'LinearSample', equal_var: bool = False) -> NamedTuple:
        """

        Parameters
        ----------
        control
        equal_var

        Returns
        -------

        """
        return t_test_from_stats(self.mean(), self.std(), self.nobs,
                                 control.mean(), control.std(), control.nobs,
                                 equal_var=equal_var)

    def mann_whitney_u_test(self, control: 'LinearSample',
                            use_continuity=True) -> NamedTuple:
        """

        Parameters
        ----------
        control
        use_continuity

        Returns
        -------

        """
        return mann_whitney_u_test(self.fullobs, control.fullobs, use_continuity)  # TODO: support weighted samples

    def shapiro_test(self) -> NamedTuple:
        """

        Returns
        -------

        """
        return shapiro_test(self.fullobs)

    def median_test(self, control: 'LinearSample') -> NamedTuple:
        """

        Parameters
        ----------
        control

        Returns
        -------

        """
        return median_test(self.fullobs, control.fullobs)

    def levene_test(self, control: 'LinearSample') -> NamedTuple:
        """

        Parameters
        ----------
        control

        Returns
        -------

        """
        return levene_test(self.fullobs, control.fullobs)

    def mood_test(self, control: 'LinearSample') -> NamedTuple:
        """

        Parameters
        ----------
        control

        Returns
        -------

        """
        return mood_test(self.fullobs, control.fullobs)


class RatioSample(SampleBase):
    """
    RatioSample
    """

    _obs_dtype = np.dtype({
        'names': ['num', 'den'],
        'formats': ['float', 'float'],
    })

    @classmethod
    def from_numden(
            cls,
            num: np.ndarray,
            den: np.ndarray,
            weights: np.ndarray = np.empty(0, dtype='int'),
    ) -> 'RatioSample':
        """

        Parameters
        ----------
        num
        den
        weights

        Returns
        -------

        """
        obs = np.rec.fromarrays([num, den], dtype=cls._obs_dtype)
        return cls(obs, weights)

    @property
    def num(self):
        """

        Returns
        -------

        """
        return self.obs['num']

    @property
    def den(self):
        """

        Returns
        -------

        """
        return self.obs['den']

    @property
    def numsamp(self):
        """

        Returns
        -------

        """
        return LinearSample(self.num, self.weights)
    
    @property
    def densamp(self):
        """

        Returns
        -------

        """
        return LinearSample(self.den, self.weights)

    def linearize(self, strategy: str = 'taylor'):
        """
        Lineariztion strategy.

        Parameters
        ----------
        strategy : str, {'taylor', 'naive'}
            When 'naive' ratio is computed element-wize. 'taylor' to use Taylor's expansion at mean point.
        Returns
        -------
        out : LinearSample
        """
        if strategy == 'taylor':
            r = self.numsamp.sum() / self.densamp.sum()
            obs = (self.num - r * self.den) / self.densamp.mean() + r
        elif strategy == 'naive':
            obs = self.num / self.den
        else:
            raise ValueError("linearizarion strategy must be 'taylor' or 'naive'")
        return LinearSample(obs, self.weights)

    def compress(
            self,
            n_buckets: int,
            sort_by: str = 'den',
            stat: str = 'mean',
            weights_dist: str = 'uniform',
            reweigh: bool = False,
            inplace=False
    ):
        """
        Compress sample into ntile-buckets

        Parameters
        ----------
        n_buckets : int
            Number of buckets / unique values in the new sample.
        sort_by : str, {'num', 'den', 'taylor', 'naive'}
        stat
        weights_dist
        reweigh
        inplace

        Returns
        -------

        """
        _validate_compress_stat(stat)

        fullobs = self.fullobs
        if sort_by == 'num':
            sort_ind = np.argsort(fullobs, order=('num', 'den'))
        elif sort_by == 'den':
            sort_ind = np.argsort(fullobs, order=('den', 'num'))
        else:
            obs_to_sort = self.linearize(strategy=sort_by).fullobs
            sort_ind = np.argsort(obs_to_sort)

        fullobs = rfn.structured_to_unstructured(fullobs[sort_ind])
        compressed, weights = compress(fullobs, n_buckets, stat, weights_dist, reweigh)

        obs = rfn.unstructured_to_structured(compressed, dtype=self._obs_dtype)
        return _return_or_inplace(self, obs, weights, inplace)

    def trim(self, ltrim=0, rtrim=0, sort_by: str = 'taylor', inplace=False):
        """

        Parameters
        ----------
        ltrim
        rtrim
        sort_by
        inplace

        Returns
        -------

        """
        if sort_by == 'num':
            trim_ind, weights = argtrim(self.obs, self.weights, ltrim=ltrim, rtrim=rtrim, order=('num', 'den'))
        elif sort_by == 'den':
            trim_ind, weights = argtrim(self.obs, self.weights, ltrim=ltrim, rtrim=rtrim, order=('den', 'num'))
        else:
            obs_to_trim = self.linearize(strategy=sort_by).obs
            trim_ind, weights = argtrim(obs_to_trim, self.weights, ltrim=ltrim, rtrim=rtrim)

        trimmed = self.obs[trim_ind]
        samp = RatioSample(trimmed, weights)

        return _return_or_inplace(self, samp.obs, samp.weights, inplace)

    def ratio(self) -> np.float:
        """

        Returns
        -------

        """
        return self.stat('ratio')


def sample(
        obs: Iterable[Union[int, float]],
        den: Iterable[Union[int, float]] = None,
        weights: Iterable[Union[int, float]] = None,
):
    """
    Create a sample — instance of LinearSample or RatioSample.
    The most convenient way to calculate statistics of sample observations.

    `obs` is 1D-observations of some random variable. If `den` is specified, returns 'RatioSample'.
    `weights` are the counts of each observation. All three parameters must be of same shape.

    Parameters
    ----------
    obs : array_like
        Observations or Numerator.
    den : array_like or None
        Denominator. Same shape as obs.
    weights : array_like or None
        Weights. Same shape as obs.
    Returns
    -------
    LinearSample or RatioSample
    """
    obs = np.asanyarray(obs)
    if weights is not None:
        weights = np.asanyarray(weights, dtype='int')
    else:
        weights = np.empty(0, dtype='int')
    if den is None:
        return LinearSample(obs, weights)
    else:
        den = np.asanyarray(den)
        return RatioSample.from_numden(obs, den, weights)


Sample = sample



from abito.core import *
from abito.lib import significance
from abito.lib import stats
from abito.lib import bootstrap
from abito.lib.prepare import compress, trim

__version__ = '0.1.2'

__all__ = [
    'significance'
]



import numpy as np
import multiprocessing
from abito.lib.stats.weighted import _quantile_sorted, _sort_obs


__all__ = ['generate_bootstrap_estimates']


def _do_bootstrap_plain(obs, stat_func, stat_args, n_iters, seed):
    if seed is not None:
        np.random.seed(seed)

    nobs = obs.shape[0]
    result = []
    for i in range(n_iters):
        new_ind = np.random.choice(nobs, nobs, replace=True)
        obs_new = obs[new_ind]
        result.append(stat_func(obs_new, **stat_args))
    return result


def _do_bootstrap_weighted(obs, weights, stat_func, stat_args, n_iters, seed):
    if seed is not None:
        np.random.seed(seed)

    if stat_func.__name__ == 'quantile':
        obs, weights = _sort_obs(obs, weights)
        stat_func = _quantile_sorted

    nobs = weights.sum()
    ps = weights / nobs

    result = []
    for i in range(n_iters):
        new_weights = np.random.multinomial(nobs, ps)
        weights = new_weights
        result.append(stat_func(weights=weights, obs=obs, **stat_args))
    return result


def _prepare_bootstrap_procedure(obs, weights, stat_func, n_iters, **stat_args):
    if weights.shape[0] == 0:
        func = _do_bootstrap_plain
        args = (obs, stat_func, stat_args, n_iters)
    else:
        func = _do_bootstrap_weighted
        args = (obs, weights, stat_func, stat_args, n_iters)
    return func, args


def generate_bootstrap_estimates(obs, stat_func, n_iters, weights=np.empty(0), n_threads=1, **stat_args):
    n_threads = multiprocessing.cpu_count() if n_threads == -1 else n_threads
    if n_threads <= 1:
        func, args = _prepare_bootstrap_procedure(obs, weights, stat_func, n_iters, **stat_args)
        results = np.asarray(func(*args, 0))
    else:
        with multiprocessing.Pool(n_threads) as pool:
            n_iters_per_thread = int(n_iters / n_threads)
            pool_results = []
            seeds = np.random.randint(2**32, size=n_threads)
            for seed in seeds:
                func, args = _prepare_bootstrap_procedure(obs, weights, stat_func, n_iters_per_thread, **stat_args)
                r = pool.apply_async(func, (*args, seed))
                pool_results.append(r)
            results = []
            [results.extend(r.get()) for r in pool_results]
            results = np.asarray(results)
    return results



import numpy as np
from typing import Union


__all__ = ['sum', 'mean', 'var', 'std', 'mean_std', 'quantile', 'median', 'ratio']


def sum(obs: np.ndarray) -> np.float:
    return obs.sum(axis=0)


def mean(obs: np.ndarray) -> np.float:
    return np.divide(obs.sum(axis=0), obs.shape[0])


def demeaned(obs: np.ndarray) -> np.ndarray:
    return obs - mean(obs)


def demeaned_sumsquares(obs: np.ndarray) -> np.float:
    return (demeaned(obs) ** 2).sum(axis=0)


def var(obs: np.ndarray) -> np.float:
    return demeaned_sumsquares(obs) / (obs.shape[0] - 1)


def std(obs: np.ndarray) -> np.float:
    return np.sqrt(var(obs))


def mean_std(obs: np.ndarray) -> np.float:
    return std(obs) / np.sqrt(obs.shape[0])


def quantile(obs: np.ndarray, q: float) -> Union[np.ndarray, np.float]:
    return np.quantile(obs, q, axis=0)


def median(obs: np.ndarray) -> np.float:
    return quantile(obs, 0.5)


def ratio(obs: np.ndarray) -> np.float:
    return sum(obs['num']) / sum(obs['den'])



import numpy as np
from typing import Iterable, Tuple
from collections import namedtuple
import scipy.stats as stats
from abito.lib.stats.plain import *


__all__ = [
    't_test_from_stats',
    't_test',
    't_test_1samp',
    'mann_whitney_u_test_from_stats',
    'mann_whitney_u_test',
    'bootstrap_test',
    'shapiro_test',
    'median_test',
    'levene_test',
    'mood_test',
]


def _unequal_var_ttest_denom(v1, n1, v2, n2):
    vn1 = v1 / n1
    vn2 = v2 / n2
    with np.errstate(divide='ignore', invalid='ignore'):
        df = (vn1 + vn2)**2 / (vn1**2 / (n1 - 1) + vn2**2 / (n2 - 1))

    # If df is undefined, variances are zero (assumes n1 > 0 & n2 > 0).
    # Hence it doesn't matter what df is as long as it's not np.nan.
    df = np.where(np.isnan(df), 1, df)
    denom = np.sqrt(vn1 + vn2)
    return df, denom


def _equal_var_ttest_denom(v1, n1, v2, n2):
    df = n1 + n2 - 2.0
    svar = ((n1 - 1) * v1 + (n2 - 1) * v2) / df
    denom = np.sqrt(svar * (1.0 / n1 + 1.0 / n2))
    return df, denom


TTestResult = namedtuple('TTestResult', [
    'statistic',
    'p_value',
    'mean_diff',
    'mean_diff_std',
])
TTestResult.__new__.__defaults__ = (np.nan,) * len(TTestResult._fields)


def _t_test_from_stats(mean1, mean2, denom, df) -> TTestResult:
    d = mean1 - mean2
    with np.errstate(divide='ignore', invalid='ignore'):
        t = np.divide(d, denom)
    prob = stats.distributions.t.sf(np.abs(t), df) * 2  # use np.abs to get upper tail

    return TTestResult(p_value=prob, statistic=t, mean_diff=d, mean_diff_std=denom)


def t_test_from_stats(mean1: float, std1: float, nobs1: int, mean2: float, std2: float, nobs2: float,
                      equal_var: bool = False) -> TTestResult:
    """
    :return: TTestResult(statistic, p_value, mean_diff, mean_diff_std)
    """
    if equal_var:
        df, denom = _equal_var_ttest_denom(std1**2, nobs1, std2**2, nobs2)
    else:
        df, denom = _unequal_var_ttest_denom(std1**2, nobs1,
                                             std2**2, nobs2)

    return _t_test_from_stats(mean1, mean2, denom, df)


def t_test(obs, obs_control, equal_var=False) -> TTestResult:
    mean1 = mean(obs)
    std1 = std(obs)
    nobs1 = obs.shape[0]
    mean2 = mean(obs_control)
    std2 = std(obs_control)
    nobs2 = obs_control.shape[0]
    return t_test_from_stats(mean1, std1, nobs1, mean2, std2, nobs2, equal_var)


def t_test_1samp(obs, popmean: float) -> TTestResult:
    """
    :param popmean: population mean
    :return: TTestResult(statistic, p_value, mean_diff, mean_diff_std)
    """
    return _t_test_from_stats(mean(obs), popmean, mean_std(obs), obs.shape[0] - 1)


def _mann_whitney_u_statistic(obs, obs_control, use_continuity: bool = True) -> Tuple:

    obs = np.asarray(obs)
    obs_control = np.asarray(obs_control)
    n1 = len(obs)
    n2 = len(obs_control)
    ranked = stats.rankdata(np.concatenate((obs, obs_control)))
    rankx = ranked[0:n1]  # get the x-ranks
    u1 = n1 * n2 + (n1 * (n1 + 1)) / 2.0 - np.sum(rankx, axis=0)  # calc U for x

    t = stats.tiecorrect(ranked)
    std = np.sqrt(t * n1 * n2 * (n1 + n2 + 1) / 12.0)

    mean = n1 * n2 / 2.0 + 0.5 * use_continuity

    return u1, mean, std


MannWhitneyUTestResult = namedtuple('MannWhitneyUTestResult', [
    'u_statistic',
    'z_statistic',
    'p_value',
])
MannWhitneyUTestResult.__new__.__defaults__ = (np.nan,) * len(MannWhitneyUTestResult._fields)


def mann_whitney_u_test_from_stats(u1: float, mean: float, std: float) -> MannWhitneyUTestResult:
    """
    :return: MannWhitneyUTestResult(u_statistic, z_statistic, p_value)
    """
    z = np.divide(-(u1 - mean), std)
    p = 2 * stats.distributions.norm.sf(abs(z))
    return MannWhitneyUTestResult(p_value=p, u_statistic=u1, z_statistic=z)


def mann_whitney_u_test(obs: Iterable, obs_control: Iterable, use_continuity: bool = True) -> MannWhitneyUTestResult:
    """
    :param use_continuity: use continuity correction
    :return: MannWhitneyUTestResult(u_statistic, z_statistic, p_value)
    """
    stats = _mann_whitney_u_statistic(obs, obs_control, use_continuity)
    return mann_whitney_u_test_from_stats(*stats)


BootstrapTestResult = namedtuple('BootstrapTestResult', [
    'estimates_diff_std',

    'est_p_value',
    'est_t_statistic',

    'median_est_diff',
    'median_est_t_statistic',
    'median_est_t_p_value',

    'stat_diff',
    't_statistic',
    't_p_value',
])
BootstrapTestResult.__new__.__defaults__ = (np.nan,) * len(BootstrapTestResult._fields)


def bootstrap_test(
        stat_val,
        bootstrap_estimates,
        nobs,
        stat_val_control,
        bootstrap_estimates_control,
        nobs_control
) -> BootstrapTestResult:
    """
    :param stat_val: sample value of statistic in treatment group
    :param bootstrap_estimates: bootstrap estimates (10000 or so) of statistic
    :param nobs: number of observations in initial sample (needed only for degrees of freedom for t-distribution)
    :param stat_val_control: sample value of statistic in control group
    :param bootstrap_estimates_control: same as above
    :param nobs_control: same as above
    :return: BootstrapTestResult(estimates_diff_std, est_p_value, est_t_statistic, median_est_diff,
                                 median_est_t_statistic, median_est_t_p_value, stat_diff, t_statistic, t_p_value)
    """
    estimates_diff = bootstrap_estimates - bootstrap_estimates_control

    median_est_diff = np.median(estimates_diff, axis=0)
    estimates_diff_std = np.std(estimates_diff, axis=0)

    est_p_value = np.min([
        (estimates_diff <= 0).mean(axis=0),
        (estimates_diff > 0).mean(axis=0)
    ], axis=0) * 2

    df = nobs + nobs_control - 2
    est_t_statistic = stats.distributions.t.isf(est_p_value / 2, df=df) * np.sign(median_est_diff)

    median_est_t_statistic = np.divide(median_est_diff, estimates_diff_std)
    median_est_t_p_value = stats.distributions.t.sf(np.abs(median_est_t_statistic), df) * 2

    stat_diff = stat_val - stat_val_control

    t_statistic = np.divide(stat_diff, estimates_diff_std)
    t_p_value = stats.distributions.t.sf(np.abs(t_statistic), df) * 2  # use np.abs to get upper tail

    return BootstrapTestResult(estimates_diff_std, est_p_value, est_t_statistic, median_est_diff,
                               median_est_t_statistic, median_est_t_p_value, stat_diff, t_statistic, t_p_value)


ShapiroTestResult = namedtuple('ShapiroTestResult', [
    'statistic',
    'p_value',
])
ShapiroTestResult.__new__.__defaults__ = (np.nan,) * len(ShapiroTestResult._fields)


def shapiro_test(obs) -> ShapiroTestResult:
    """
    Shapiro-Wilk test of normality
    :return: ShapiroTestResult(statistic, p_value)
    """
    if obs.size < 3:
        return ShapiroTestResult(np.nan, np.nan)
    res = stats.shapiro(obs)
    return ShapiroTestResult(p_value=res[1], statistic=res[0])


MedianTestResult = namedtuple('MedianTestResult', [
    'statistic',
    'p_value',
    'grand_median',
])
MedianTestResult.__new__.__defaults__ = (np.nan,) * len(MedianTestResult._fields)


def median_test(obs, obs_control) -> MedianTestResult:
    """
    :return: MedianTestResult(statistic, p_value, grand_median)
    """
    if std(obs) == 0:
        return MedianTestResult(np.nan, np.nan, np.nan)
    res = stats.median_test(obs, obs_control)
    return MedianTestResult(p_value=res[1], statistic=res[0], grand_median=res[2])


LeveneTestResult = namedtuple('LeveneTestResult', [
    'statistic',
    'p_value',
])
LeveneTestResult.__new__.__defaults__ = (np.nan,) * len(LeveneTestResult._fields)


def levene_test(obs, obs_control) -> LeveneTestResult:
    """
    :return: LeveneTestResult(statistic, p_value)
    """
    res = stats.levene(obs, obs_control)
    return LeveneTestResult(p_value=res[1], statistic=res[0])


MoodTestResult = namedtuple('MoodTestResult', [
    'statistic',
    'p_value',
])
MoodTestResult.__new__.__defaults__ = (np.nan,) * len(MoodTestResult._fields)


def mood_test(obs, obs_control) -> MoodTestResult:
    """
    :return: MoodTestResult(statistic, p_value)
    """
    res = stats.mood(obs, obs_control)
    return MoodTestResult(statistic=res[0], p_value=res[1])


import numpy as np
from abito.lib.stats import plain


def _array_binary_view(ar: np.ndarray):
    ar = np.ascontiguousarray(ar)
    ar = ar.view('|S%d' % (ar.itemsize * ar[0].size))
    return ar


def _reweigh_plain_general(ar: np.ndarray):
    ar_binary = _array_binary_view(ar)
    _, ind, counts = np.unique(ar_binary, return_index=True, return_counts=True)
    return ar[ind], counts


def _reweigh_plain_int_fast(ar: np.ndarray):
    try:
        y = np.bincount(ar)
    except (ValueError, TypeError):
        ar_int = ar.astype('int', casting='unsafe')
        assert (ar == ar_int).all()
        y = np.bincount(ar_int)
    ii = np.nonzero(y)[0]
    u, counts = ii, y[ii]
    return u, counts


def _reweigh_plain(ar: np.ndarray):
    try:
        return _reweigh_plain_int_fast(ar)
    except (ValueError, TypeError, AssertionError):
        return _reweigh_plain_general(ar)


def _reweigh_weighted(ar: np.ndarray, weights: np.ndarray):
    ar_binary = _array_binary_view(ar)
    if ar_binary.ndim > 1:
        ar_binary = ar_binary.flatten()
    unique_values, ind = np.unique(ar_binary, return_index=True)
    new_weights = np.array([weights[ar_binary == e].sum() for e in unique_values])
    return ar[ind], new_weights


def reweigh(ar: np.ndarray, weights: np.ndarray = np.empty(0)):
    if weights.shape[0] == 0:
        compressed, weights = _reweigh_plain(ar)
    else:
        compressed, weights = _reweigh_weighted(ar, weights)
    return compressed, weights


def _calc_bucket_weights(nobs: int, n_buckets: int, dist: str = 'uniform') -> np.ndarray:
    if dist == 'uniform':
        nobs_per_bucket = nobs // n_buckets
        nobs_left = nobs % n_buckets
        leftovers = np.array((n_buckets - nobs_left) * [0] + nobs_left * [1])
        np.random.shuffle(leftovers)
        bucket_weights = np.array([nobs_per_bucket] * n_buckets) + leftovers
    elif dist == 'multinomial':
        bucket_weights = np.random.multinomial(nobs, [1 / n_buckets] * n_buckets)
    else:
        raise ValueError('dist must be uniform or multinomial')
    return bucket_weights.astype('int')


def _split_into_buckets(
        ar: np.ndarray,
        n_buckets: int,
        weights_dist: str = 'uniform',
):
    nobs = ar.shape[0]
    bucket_weights = _calc_bucket_weights(nobs, n_buckets, weights_dist)
    bucket_split_indices = bucket_weights.cumsum()[:-1]

    return np.split(ar, bucket_split_indices), bucket_weights


def _validate_compress_stat(stat):
    if stat not in ('mean', 'median'):
        raise ValueError('Statistic for compression can only be mean or median')


def compress(
        ar: np.ndarray,
        n_buckets: int,
        stat,
        weights_dist: str = 'uniform',
        reweigh: bool = False,
):
    split, weights = _split_into_buckets(ar, n_buckets, weights_dist)

    stat_func = getattr(plain, stat)
    compressed = np.array([stat_func(s) for s in split])

    if reweigh:
        return _reweigh_weighted(compressed, weights)
    else:
        return compressed, weights



from typing import Tuple, Union
import numpy as np
from abito.lib.utils import _quantile_is_valid


def _argquantile_weighted(
        weights_sorted: np.ndarray,
        q: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    inds = []
    cuts = []
    q = np.asanyarray(q)
    weight_qs = [int(q * weights_sorted.sum()) for q in q.flatten()]
    cur_weight_q = 0
    weight_q = weight_qs.pop(0)
    for cur_weight_ind, cur_weight in enumerate(weights_sorted):
        cur_weight_q += cur_weight
        while weight_q <= cur_weight_q:
            inds.append(cur_weight_ind)
            cuts.append(cur_weight - (cur_weight_q - weight_q))
            if weight_qs:
                weight_q = weight_qs.pop(0)
            else:
                weight_q = -1
                break
        if weight_q < 0:
            break
    inds = np.asanyarray(inds).reshape(q.shape)
    cuts = np.asanyarray(cuts).reshape(q.shape)
    return inds, cuts


def _argtrim_plain(
        ar: np.ndarray,
        q: np.ndarray,
        **kwargs
):
    trim_inds = (q * ar.shape[0]).astype('int')
    ind = np.argpartition(ar, trim_inds, **kwargs)
    return ind[trim_inds[0]:trim_inds[1]]


def _argtrim_weighted(
        ar: np.ndarray,
        weights: np.ndarray,
        q: np.ndarray,
        **kwargs
) -> Tuple[np.ndarray, Union[np.ndarray, None]]:

    weights = weights.copy()

    ind_sorted = ar.argsort(**kwargs)
    weights = weights[ind_sorted]

    (lowercut_a, uppercut_a), (lowercut_w, uppercut_w) = _argquantile_weighted(weights, q)

    weights[uppercut_a] = uppercut_w
    weights[lowercut_a] -= lowercut_w

    ind_sorted = ind_sorted[lowercut_a:uppercut_a + 1]
    weights = weights[lowercut_a:uppercut_a + 1]
    return ind_sorted, weights


def argtrim(
        ar: np.ndarray,
        weights: np.ndarray = np.empty(0),
        ltrim: float = 0,
        rtrim: float = 0,
        **kwargs
):
    q = np.array([ltrim, 1 - rtrim])
    _quantile_is_valid(q)
    if weights.shape[0] == 0:
        ind = _argtrim_plain(ar, q, **kwargs)
    else:
        ind, weights = _argtrim_weighted(ar, weights, q, **kwargs)
    return ind, weights



import pathlib
from setuptools import setup, find_packages


with open('requirements.txt') as f:
    required = f.read().splitlines()

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setup(
    name='abito',
    version='0.1.2',
    license='MIT',
    description='Package for hypothesis testing in A/B-experiments',
    long_description=README,
    long_description_content_type="text/markdown",
    author='Danila Lenkov',
    author_email='dlenkoff@gmail.com',
    url='https://github.com/avito-tech/abito',
    packages=find_packages(exclude=("tests",)),
    install_requires=required,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering'
    ]
)



import numpy as np


__all__ = ['is_array_sorted']


def is_array_sorted(ar):
    for i in range(ar.size-1):
        if ar[i + 1] < ar[i]:
            return False
    return True


def _quantile_is_valid(q):
    res = True
    # avoid expensive reductions, relevant for arrays with < O(1000) elements
    if q.ndim == 1 and q.size < 10:
        for i in range(q.size):
            if q[i] < 0.0 or q[i] > 1.0:
                res = False
    else:
        # faster than any()
        if np.count_nonzero(q < 0.0) or np.count_nonzero(q > 1.0):
            res = False
    if not res:
        raise ValueError("Quantiles must be in the range [0, 1]")
    return True


def _return_or_inplace(obj, obs, weights, inplace):
    if inplace:
        obj.__init__(obs, weights)
    else:
        return obj.__class__(obs, weights)



import numpy as np
from typing import Union
from abito.lib.utils import is_array_sorted, _quantile_is_valid
from abito.lib.prepare.trim import _argquantile_weighted


__all__ = ['sum', 'mean', 'var', 'std', 'mean_std', 'quantile', 'median', 'ratio']


def sum(obs: np.ndarray, weights: np.ndarray) -> np.float:
    return np.dot(obs.T, weights).sum()


def mean(obs: np.ndarray, weights: np.ndarray) -> np.float:
    return np.divide(sum(obs, weights), weights.sum())


def demeaned(obs: np.ndarray, weights: np.ndarray) -> np.ndarray:
    return obs - mean(obs, weights)


def demeaned_sumsquares(obs: np.ndarray, weights: np.ndarray) -> np.float:
    return np.dot((demeaned(obs, weights) ** 2).T, weights).sum()


def var(obs: np.ndarray, weights: np.ndarray) -> np.float:
    return demeaned_sumsquares(obs, weights) / (weights.sum() - 1)


def std(obs: np.ndarray, weights: np.ndarray) -> np.float:
    return np.sqrt(var(obs, weights))


def mean_std(obs: np.ndarray, weights: np.ndarray):
    return std(obs, weights) / np.sqrt(weights.sum())


def ratio(obs: np.ndarray, weights: np.ndarray) -> np.float:
    return sum(obs['num'], weights) / sum(obs['den'], weights)


def _quantile_sorted(obs: np.ndarray, weights: np.ndarray, q) -> Union[np.ndarray, np.float]:
    indices, _ = _argquantile_weighted(weights, q)
    return obs[indices]


def _sort_obs(obs: np.ndarray, weights: np.ndarray):
    if not is_array_sorted(obs):
        ind_sorted = np.argsort(obs)
        obs = obs[ind_sorted]
        weights = weights[ind_sorted]
    return obs, weights


def quantile(obs: np.ndarray, weights: np.ndarray, q) -> Union[np.ndarray, np.float]:
    obs = np.asarray(obs)
    weights = np.asarray(weights)
    q = np.asanyarray(q)
    _quantile_is_valid(q)
    obs, weights = _sort_obs(obs, weights)
    return _quantile_sorted(obs, weights, q)


def median(obs: np.ndarray, weights: np.ndarray) -> np.float:
    return quantile(obs, weights, 0.5)


